"""

A "middleware" is a function that works with every request before it
is processed by any specific path operation. And also with every response before returning it.

It takes each request that comes to your application.
It can then do something to that request or run any needed code.
Then it passes the request to be processed by the rest of the application (by some path operation).
It then takes the response generated by the application (by some path operation).
It can do something to that response or run any needed code.
Then it returns the response.


To create a middleware you use the decorator @app.middleware("http") on top of a function.

The middleware function receives:

The request.
A function call_next that will receive the request as a parameter.
This function will pass the request to the corresponding path operation.
Then it returns the response generated by the corresponding path operation.
You can then modify further the response before returning it.

"""


from fastapi import FastAPI, APIRouter, Request
import uvicorn
import time

router = APIRouter()
app = FastAPI()


@app.middleware("http")
async def log_request_time(request: Request, call_next):
    start_time = time.time()
    print(f'Start time: {start_time}')
    response = await call_next(request)
    process_time = time.time() - start_time
    response.headers['x-Process-Time'] = str(process_time)
    print(f'process_time: {process_time}')
    return response


@router.get("/users")
async def get_users():
    print('Getting users....')
    return [{"id": 1, "name": "user1"}]


@router.get("/products")
async def get_users():
    print('Getting products....')
    return [{"id": 1, "name": "p1"}]


if __name__ == '__main__':
   app.include_router(router=router, prefix="/api/v1", tags=["Test Middleware"])
   uvicorn.run(app)


"""
C:\Python3\python.exe "C:/Users/aafakmoh/OneDrive - Hewlett Packard Enterprise/mypy/app/fast_api_exp/middleware_exp.py"
INFO:     Started server process [19300]
INFO:     Waiting for application startup.
INFO:     Application startup complete.
INFO:     Uvicorn running on http://127.0.0.1:8000 (Press CTRL+C to quit)
Start time: 1663136812.2432935
Getting users....
process_time: 0.0030074119567871094
INFO:     127.0.0.1:56262 - "GET /api/v1/users HTTP/1.1" 200 OK


Repsonse Headers:
date:	Wed, 14 Sep 2022 06:26:52 GMT
server:	uvicorn
content-length:	25 bytes
content-type:	application/json
x-process-time:	0.0030074119567871094
"""